# KungFu - A GStreamer based DVD ripper
# Copyright (C) 2006 Jason Gerard DeRose <jderose@jasonderose.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import time
import gst
import gtk
import gobject
import cairo
from kf_constants import *
import kf_state
from kf_obj import Title
from kf_util import Timer, format_time, format_posdur, Paths
import kf_info

import logging
log = logging.getLogger('kf_gst')

# See if dvdsrc is available
try:
	gst.element_factory_make('dvdsrc')
	DVDSRC = 'dvdsrc'
except gst.PluginNotFoundError: 
	DVDSRC = 'dvdreadsrc'
log.info('Using %s element' % DVDSRC)


class Base(gobject.GObject):
	"""
	Base class for GStreamer backend.  Provides simple API for higher level
	GUI components.  Takes care of updating stream position, duration,
	elapsed time, etc.
	"""

	__gsignals__ = {
		'playing': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_BOOLEAN]),
		'update_time': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_FLOAT, gobject.TYPE_FLOAT]),
		'update_stats': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_FLOAT, gobject.TYPE_FLOAT, gobject.TYPE_FLOAT]),
		'eos': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
	} # __gsignals__
	
	def __init__(self, playbin=False):
		gobject.GObject.__init__(self)
		self._playing = False
		self._start = 0.0
		self._elapsed = 0.0
		self._timer = Timer(1, self._on_timer)
		if playbin:
			self._pipeline = gst.element_factory_make('playbin')
		else:
			self._pipeline = gst.Pipeline()
		self._bus = self._pipeline.get_bus()
		self._bus.add_signal_watch()
		self._bus.connect('message::eos', self._on_eos)
		self._bus.connect('message::error', self._on_error)
			
	def _on_eos(self, bus, msg):
		log.info('eos')
		self.stop()
		self.emit('eos')
				
	def _on_error(self, bus, msg):
		error = msg.parse_error()
		log.error(error[1])
	
	def _on_timer(self):
		pos = self.get_pos()
		dur = self.get_dur()
		self.emit('update_time', pos, dur)
		elapsed = self._elapsed + time.time() - self._start
		if elapsed > 0:
			rate = pos / elapsed
		else:
			rate = 0.0
		if rate > 0:
			remaining = (dur - pos) / rate
		else:
			remaining = 0.0
		self.emit('update_stats', rate, elapsed, remaining)
	
	def _set_playing(self, playing):
		self._playing = playing
		self.emit('playing', playing)
		if playing:
			self._timer.start()
		else:
			self._timer.stop()
			
	def _on_idle(self, callback):
		callback()
		return False # Don't repeat idle call
		
	def _seek_idle(self, sec):
		ns = long(sec * gst.SECOND)
		self._pipeline.seek_simple(
			gst.FORMAT_TIME, 
			gst.SEEK_FLAG_FLUSH,
			ns
		)
		self._timer.restart() # Only starts if it was already running
		return False # Don't repeat idle call

	# Public Methods
	def play_pause(self):
		if self._playing:
			self.pause()
		else:
			self.play()

	def play(self):
		self._pipeline.set_state(gst.STATE_PLAYING)
		self._start = time.time()
		self._set_playing(True)
	
	def pause(self):
		self._pipeline.set_state(gst.STATE_PAUSED)
		self._elapsed += (time.time() - self._start)
		self._set_playing(False)

	def stop(self):
		self._pipeline.set_state(gst.STATE_READY)
		self._start = 0.0
		self._elapsed = 0.0
		self._set_playing(False)
		
	def shutdown(self):
		'''Complete shutdown without causing further signal emision'''
		self._pipeline.set_state(gst.STATE_NULL)
		self._timer.stop()
		
	def get_pos(self):
		(pos, format) = self._pipeline.query_position(gst.FORMAT_TIME)
		return pos/float(gst.SECOND)
		
	def get_dur(self):
		(dur, format) = self._pipeline.query_duration(gst.FORMAT_TIME)
		return dur/float(gst.SECOND)
		
	def seek(self, sec):
		# element.seek() isn't thread-safe, and it's called from outside the
		# main thread at least in TitleReader, thus the idle call:
		gobject.idle_add(self._seek_idle, sec)
		
	def mainthread(self, callback):
		gobject.idle_add(self._on_idle, callback)
		

		
class TitleReader(Base):
	def __init__(self, device, title, callback):
		Base.__init__(self)
		self._title = title
		self._callback = callback
		self._first_preroll = True
		
		# Create elements
		src = gst.element_factory_make(DVDSRC)
		demux = gst.element_factory_make('dvddemux')
		dec = gst.element_factory_make('mpeg2dec')
		deint = gst.element_factory_make('ffdeinterlace')
		cspace = gst.element_factory_make('ffmpegcolorspace')
		enc = gst.element_factory_make('pngenc')
		sink = gst.element_factory_make('fakesink')
		
		# Add to pipeline
		self._pipeline.add(src, demux, dec, deint, cspace, enc, sink)

		# Set properties
		src.set_property('device', device)
		src.set_property('title', title)
		sink.set_property('signal-handoffs', False)
		
				
		# Connect handlers		
		dec.get_pad('src').connect('notify::caps', self._on_notify_caps)
		sink.connect('preroll-handoff', self._on_preroll_handoff)

		# Link elements
		gst.element_link_many(src, demux, dec, deint, cspace, enc, sink)
		
		self._pipeline.set_state(gst.STATE_PAUSED)
		self._pipeline.get_state()
		self._dur = self.get_dur()
		
		if self._dur < MIN_DURATION:
			log.warn('Skipping title %d: %d <  %d (MIN_DURATION)', self._title, self._dur, MIN_DURATION)
			self.shutdown()
			self.mainthread(lambda: self._callback(None, None, None, None))
		else:
			log.info('Title %d has duration %s', self._title, format_time(self._dur))
			sink.set_property('signal-handoffs', True)
			self.seek(self._dur * 0.25)

	def _on_notify_caps(self, pad, args):
		caps = pad.get_negotiated_caps()
		if caps == None:
			return
		self._width = caps[0]['width']
		self._height = caps[0]['height']
		frac = caps[0]['pixel-aspect-ratio']
		self._aspect = frac.num / float(frac.denom)
		log.info('Title %d is %dx%d with %d:%d pixel-aspect-ratio', self._title, self._width, self._height, frac.num, frac.denom)

	def _on_preroll_handoff(self, sink, buf, pad):
		print 'preroll', self._title
		png = Paths.get_tmpfile('%02d.png' % self._title)
		f = open(png, 'w')
		f.write(str(buf))
		f.close()
		if self._aspect > 1:
			w = int(self._width * self._aspect)
			h = self._height
		else:
			w = self._width
			h = int(self._height / self._aspect)
		# HACK: 
		# I'm not aspect correcting in the gst pipeline because of
		# gstreamer bug #364856. gdk.pixbuf_new_from_file_at_size() 
		# wont change the aspect ratio, so I must do it in the PNG or it will
		# be wrong in the gtk.CellRendererPixbuf
		# As a side benifit, cairo does very pretty scaling.  :)
		src = cairo.ImageSurface.create_from_png(png)
		dst = cairo.ImageSurface(cairo.FORMAT_RGB24, w, h)
		ctx = cairo.Context(dst)
		sx = w / float(self._width)
		sy = h / float(self._height)
		ctx.scale(sx, sy)
		ctx.set_source_surface(src, 0, 0)
		ctx.paint()
		dst.write_to_png(png)
		# /HACK
		self.mainthread(lambda: self._callback(self._dur, w, h, png))
		self.stop()



class Reader(Base):
	'''Reads title information from DVD'''
	
	__gsignals__ = {
		'reading': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_BOOLEAN]),
	} # __gsignals__
	
	def __init__(self):
		Base.__init__(self)
		self._titlereader = None
		self.model = gtk.ListStore(object, bool, str, str, gtk.gdk.Pixbuf)
	
	def clear(self):
		self.shutdown()
		self.model.clear()
		self.emit('reading', False)

	def read(self, source):
		self.emit('reading', True)
		self._device = source.device
		self._prefix = source.prefix
		self.model.clear()
		log.info('Reading titles from %s at %s', self._prefix, self._device)
		self._read_num_title_list()
		self._read_title()
	
	def shutdown(self):
		if self._titlereader != None:
			self._titlereader.shutdown()
	
	def _read_title(self):
		if self._i < self._titles:
			self._titlereader = TitleReader(self._device, self._i + 1, self._titlereader_finished)
		else:
			self._titlereader = None
			self.emit('reading', False)	
	
	def _read_num_title_list(self):
		pipeline = gst.Pipeline()
		src = gst.element_factory_make(DVDSRC)
		src.set_property('device', self._device)
		sink = gst.element_factory_make('fakesink')
		pipeline.add(src, sink)
		src.link(sink)
		pipeline.set_state(gst.STATE_PAUSED)
		pipeline.get_state(1000)
		(self._titles, format) = pipeline.query_duration(
			gst.format_get_by_nick('title'))
		self._i = 0
		log.info('%d titles found on %s', self._titles, self._prefix)
	
	def _titlereader_finished(self, dur, width, height, png):
		if dur != None:
			title =  Title(self.model, len(self.model), self._device, self._i+1, self._prefix)
			title.dur = dur
			title.width = width
			title.height = height
			title.png = png 

			self.model.append([
				title, 
				False, 
				str(title.title),
				format_time(title.dur),
				gtk.gdk.pixbuf_new_from_file_at_size(png, width / 6, height / 6),
			])
		self._i += 1
		self._read_title()



class AudioTranscoder(gst.Bin):
	"""
	Complete audio transcoder.
	Decodes, converts, resamples, and lastly encodes to vorbis.
	"""
	
	def __init__(self, title):
		gst.Bin.__init__(self)
		self._title = title
		
		# Create elements
		self._inq = gst.element_factory_make('queue')
		self._conv = gst.element_factory_make('audioconvert')
		self._rsmpl = gst.element_factory_make('audioresample')
		self._enc = gst.element_factory_make('vorbisenc')
		self._outq = gst.element_factory_make('queue')
		
		# Set properties
		self._enc.set_property('quality', title.get('vorbis_quality'))
		
		log.info('Vorbis quality = %.1f', self._enc.get_property('quality'))
		
		# Add to bin
		self.add(self._inq, self._conv, self._rsmpl, self._enc, self._outq)
		
		# Connect 'notify::caps' handlers
		self._inq.get_pad('sink').connect(
			'notify::caps', self._on_notify_caps1)
		self._conv.get_pad('sink').connect(
			'notify::caps', self._on_notify_caps2)
		
		# Link *some* elements
		# Completed in _on_notify_caps1(), _on_notify_caps2()
		self._conv.link(self._rsmpl, 
			gst.caps_from_string('audio/x-raw-float, channels=2'))
		self._enc.link(self._outq)
		
		# Add ghostpads
		self.add_pad(gst.GhostPad('sink', self._inq.get_pad('sink')))
		self.add_pad(gst.GhostPad('src', self._outq.get_pad('src')))
	
	def _on_notify_caps1(self, pad, args):
		caps = pad.get_negotiated_caps()
		if caps == None or self._inq.get_pad('src').is_linked():
			return
		name = caps[0].get_name()
		if name not in AUDIODEC:
			log.error('No mapping for audio type %s', name)
			return
		log.info('Audio type is %s, decoding with %s', name, AUDIODEC[name])
		self._dec = gst.element_factory_make(AUDIODEC[name])
		self.add(self._dec)
		gst.element_link_many(self._inq, self._dec, self._conv)
		self._dec.set_state(gst.STATE_PLAYING)
	
	def _on_notify_caps2(self, pad, args):
		# DVD audio sample rate should be 48000 or 96000... even though
		# vorbisenc supports up to 50000, we force down sampling to 48000
		# as common hardware doesn't support above 48000.
		# Personally, I disagree with Thoggen's downsampling to 44100.
		# It slows the transcoding a bit, causes some quality loss, and saves
		# very little space.  In a test I just did, downsampling a
		# 48000 stream to 44100 (encoded with quality=0.5) only reduced
		# the audio-only file size from 11,776,951 bytes to 11,685,782 bytes.
		# If the user's soundcard doesn't support 48000... well, he/she
		# would have the same problem playing the DVD as playing the rip,
		# i.e., the audio must be downsampled during playback.
		# However, if audio is at some non-standard sample rate below 48000,
		# we don't force upsampling to 48000... we use min(src_rate, 48000).
		caps = pad.get_negotiated_caps()
		if caps == None or self._rsmpl.get_pad('src').is_linked():
			return
		src_rate = caps[0]['rate']
		rate = min(src_rate, 48000)
		log.info('Audio sample rate is %d, output will be %d', src_rate, rate) 
		self._rsmpl.link(self._enc,
		 gst.caps_from_string('audio/x-raw-float, rate=%d' % rate))



class VideoTranscoder(gst.Bin):
	"""
	Complete video transcoder.
	Decodes, deinterlaces, scales, and lastly encodes to theora.
	"""

	def __init__(self, title):
		gst.Bin.__init__(self)
		
		self._size = title.get('video_size')
		
		# Create elements
		self._inq = gst.element_factory_make('queue')
		self._dec = gst.element_factory_make('mpeg2dec')
		self._deint = gst.element_factory_make('ffdeinterlace')
		self._scale = gst.element_factory_make('videoscale')
		self._rate = gst.element_factory_make('videorate')
		self._enc = gst.element_factory_make('theoraenc')
		self._outq = gst.element_factory_make('queue')
		
		# Set properties
		self._scale.set_property('method', 1) # Use Bilinear scaling
		self._enc.set_property('quality', title.get('theora_quality'))
		self._enc.set_property('quick', THEORA_QUICK)
		self._enc.set_property('sharpness', THEORA_SHARPNESS)
		
		log.info('Theora quality = %d', self._enc.get_property('quality'))
		log.info('Theora quick = %s', self._enc.get_property('quick'))
		log.info('Theora sharpness = %d', self._enc.get_property('sharpness'))
		
		# Add to bin
		self.add(self._inq, self._dec, self._deint, self._scale, 
			self._rate, self._enc, self._outq)
		
		# Connect 'notify::caps' handler
		pad = self._dec.get_pad('src')
		pad.connect('notify::caps', self._on_notify_caps)
			
		# Link *some* elements
		# Completed in _on_notify_caps()
		gst.element_link_many(self._inq, self._dec, self._deint, self._scale)
		gst.element_link_many(self._rate, self._enc, self._outq)
		
		# Add ghostpads
		self.add_pad(gst.GhostPad('sink', self._inq.get_pad('sink')))
		self.add_pad(gst.GhostPad('src', self._outq.get_pad('src')))

	def _on_notify_caps(self, pad, args):
		caps = pad.get_negotiated_caps()
		if caps == None or self._scale.get_pad('src').is_linked():
			return
		(n, d) = VIDEOSIZE[self._size]
		src_width = caps[0]['width']
		src_height = caps[0]['height']
		width = src_width * n / d
		height = src_height * n / d
		log.info('Video is %dx%d, output will be %dx%d',
			src_width, src_height, width, height)
		self._scale.link(self._rate,
		 gst.caps_from_string('video/x-raw-yuv, width=%d, height=%d'
		 % (width, height)))



class TitleRipper(Base):
	def __init__(self, title):
		Base.__init__(self)
		self.title = title
			
		# Create elements
		self._src = gst.element_factory_make(DVDSRC)
		self._demux = gst.element_factory_make('dvddemux')
		self._audio = AudioTranscoder(title)
		self._video = VideoTranscoder(title)
		self._mux = gst.element_factory_make('oggmux')
		self._sink = gst.element_factory_make('filesink')
		
		# Set properties
		self._src.set_property('device', title.device)
		self._src.set_property('title', title.title)
		self._sink.set_property('location', title.outfile)
		
		# Add elements to pipeline
		self._pipeline.add(self._src, self._demux, self._audio, self._video,
			self._mux, self._sink)
			
		# Link elements
		self._src.link(self._demux)
		self._demux.get_pad('current_audio').link(self._audio.get_pad('sink'))
		self._demux.get_pad('current_video').link(self._video.get_pad('sink'))
		self._audio.link(self._mux)
		self._video.link(self._mux)
		self._mux.link(self._sink)



class Ripper(gobject.GObject):
	__gsignals__ = {
		'ripping': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_BOOLEAN]),
		'ripping_title': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_BOOLEAN]),
		'title_progress': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_FLOAT, gobject.TYPE_GSTRING]),
		'total_progress': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_FLOAT, gobject.TYPE_GSTRING]),
		'update_stats': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_FLOAT, gobject.TYPE_FLOAT, gobject.TYPE_FLOAT]),
		'title_changed': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_PYOBJECT]),
	} # __gsignals__
	
	def __init__(self):
		gobject.GObject.__init__(self)
		self._titleripper = None
		self._ripping = False
		
	def rip(self, title_list):
		self.emit('ripping', True)
		self._title_list = title_list
		self._i = 0	
		self._pos = 0
		self._dur = 0
		for title in title_list:
			self._dur += title.dur
		self._rip_title()
		
	def cancel(self):
		self.shutdown()
		self.emit('ripping', False)
	
	def pause_resume(self):
		self._ripping = not self._ripping
		if self._titleripper != None:
			if self._ripping:
				self._titleripper.play()
			else:
				self._titleripper.pause()
		return self._ripping
	
	def _rip_title(self):
		if self._i < len(self._title_list):
			title = self._title_list[self._i]
			log.info('Ripping title %d from %s at %s to %s', 
				title.title, title.prefix, title.device, title.outfile)
			self.emit('title_changed', title)
			self._titleripper = TitleRipper(title)
			self._titleripper.connect('eos', self._titleripper_finished)
			self._titleripper.connect('update_time', self._on_update_time)
			self._titleripper.connect('update_stats', self._on_update_stats)
			self._titleripper.play()
			self._ripping = True
		else:
			self.shutdown()
			self.emit('ripping', False)		

	def _titleripper_finished(self, titleripper):
		title = titleripper.title
		title.rip = False
		log.info('Finished ripping title %d from %s at %s to %s', 
				title.title, title.prefix, title.device, title.outfile)
		self._pos += title.dur
		self._i += 1
		self._rip_title()
		
	def _on_update_time(self, titleripper, pos, dur):
		if dur > 0:
			fraction = pos / dur
		else:
			fraction = 0.0
			if fraction > 1.0:
				fraction = 1.0
			elif fraction < 0.0:
				fraction = 0.0
		text = format_posdur(pos, dur)
		self.emit('title_progress', fraction, text)
		
		if self._dur > 0:
			fraction = (pos + self._pos) / self._dur
			if fraction > 1.0:
				fraction = 1.0
			elif fraction < 0.0:
				fraction = 0.0
		else:
			fraction = 0.0
		text = format_posdur(pos + self._pos, self._dur)
		self.emit('total_progress', fraction, text)
		
	def _on_update_stats(self, titleripper, rate, elapsed, eta):
		self.emit('update_stats', rate, elapsed, eta)
	
	def shutdown(self):
		if self._titleripper != None:
			self._titleripper.shutdown()
			self._titleripper = None
			
			
		
		
class Player(Base):
	__gsignals__ = {
		'imagesink': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
			[gobject.TYPE_OBJECT]),
	} # __gsignals__
	
	def __init__(self):
		Base.__init__(self, playbin=True)
		self._bus.enable_sync_message_emission()
		self._bus.connect('sync-message::element', self._on_sync_message)
		self._pipeline.connect('notify::source', self._on_source)

	def set_title(self, device, title):
		self._device = device
		self._title = title
		self.stop()
		self._pipeline.set_property('uri', 'dvd://%d' % title)

	def _on_sync_message(self, bus, message):
		if message.structure is None:
			return
		if message.structure.get_name() == 'prepare-xwindow-id':
			message.src.set_property('force-aspect-ratio', True)
			self.emit('imagesink', message.src)

	def _on_source(self, playbin, gparam):
		src = playbin.get_by_name(gparam.name)
		src.set_property('device', self._device)
	


class GstBackend:
	def __init__(self):
		self.reader = Reader()
		self.player = Player()
		self.ripper = Ripper()
		
	def shutdown(self):
		self.reader.shutdown()
		self.player.shutdown()
		self.ripper.shutdown()
		
	def stop(self):
		self.reader.stop()
		self.player.stop()
		self.ripper.stop()
